<?php
/************************************************************************
 * Mailer v0.2.1-FINAL                                Start: 08/29/2004 *
 * ===================                          Last change: 08/29/2004 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : lib-mysql3.php                                   *
 * -------------------------------------------------------------------- *
 * Short description : Database layer for MySQL 3/4/5 server            *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : Datenbankschicht fuer MySQL 3/4/5 Server         *
 * -------------------------------------------------------------------- *
 * $Revision::                                                        $ *
 * $Date::                                                            $ *
 * $Tag:: 0.2.1-FINAL                                                 $ *
 * $Author::                                                          $ *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2009 by Roland Haeder                           *
 * Copyright (c) 2009 - 2012 by Mailer Developer Team                   *
 * For more information visit: http://mxchange.org                      *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if (!defined('__SECURITY')) {
	die();
} // END - if

// SQL queries
function SQL_QUERY ($sqlString, $F, $L, $enableCodes = TRUE) {
	// Is there cache?
	if (!isset($GLOBALS[__FUNCTION__][$sqlString])) {
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Called: ' . $sqlString);

		// Trim SQL string
		$sqlStringModified = trim($sqlString);

		// Empty query string or link is not up?
		if (empty($sqlStringModified)) {
			// Empty SQL string!
			reportBug(__FUNCTION__, __LINE__, sprintf("SQL string is empty, please fix this: file=%s, line=%s",
				basename($F),
				$L
			));
		} elseif (!SQL_IS_LINK_UP()) {
			// We should not quietly ignore this
			reportBug(__FUNCTION__, __LINE__, sprintf("Cannot query database: sqlString=%s,file=%s,line=%s",
				$sqlStringModified,
				basename($F),
				$L
			));
		}

		// Remove \t, \n and \r from queries they may confuse some MySQL versions
		$sqlStringModified = str_replace(array(chr(9), chr(10), chr(13)), array(' ', ' ', ' '), $sqlStringModified);

		// Compile config entries out
		$sqlStringModified = SQL_PREPARE_SQL_STRING($sqlStringModified, $enableCodes);

		// Cache it and remember as last SQL query
		$GLOBALS[__FUNCTION__][$sqlString] = $sqlStringModified;
		$GLOBALS['last_sql']               = $sqlStringModified;
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Stored cache: ' . $sqlStringModified);
	}  else {
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Cache used: ' . $sqlString);

		// Use cache (to save a lot function calls
		$GLOBALS['last_sql'] = $GLOBALS[__FUNCTION__][$sqlString];

		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Cache is: ' . $sqlString);
	}

	// Starting time
	$querytimeBefore = microtime(TRUE);

	// Run SQL command
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'F=' . basename($F) . ',L=' . $L . ',sql=' . $GLOBALS['last_sql']);
	$result = mysql_query($GLOBALS['last_sql'], SQL_GET_LINK())
		or SQL_ERROR($F, $L, 'file='. basename($F) . ',line=' . $L . ':mysql_error()=' . mysql_error() . ',last_query=' . $GLOBALS['last_sql']);
	//* DEBUG: */ logDebugMessage($F, $L, 'sql=' . $GLOBALS['last_sql'] . ',affected=' . SQL_AFFECTEDROWS() . ',numRows='.(is_resource($result) ? SQL_NUMROWS($result) : gettype($result)));

	// Calculate query time
	$queryTime = microtime(TRUE) - $querytimeBefore;

	// Add this query to array including timing
	addSqlToDebug($result, $GLOBALS['last_sql'], $queryTime, $F, $L);

	// Save last successfull query
	setConfigEntry('db_last_query', $GLOBALS['last_sql']);

	// Count all query times
	incrementConfigEntry('sql_time', $queryTime);

	// Count this query
	incrementConfigEntry('sql_count');

	// Debug output
	if (SQL_DEBUG_ENABLED()) {
		// Is this the first call?
		if (!isset($GLOBALS['sql_first_entry'])) {
			// Write first entry
			appendLineToFile(getCachePath() . 'mysql.log', 'Module=' . getModule());
			$GLOBALS['sql_first_entry'] = TRUE;
		} // END - if

		// Append debug line
		appendLineToFile(getCachePath() . 'mysql.log', basename($F) . '|LINE=' . $L . '|NUM=' . (is_resource($result) ? SQL_NUMROWS($result) : 'false') . '|AFFECTED=' . SQL_AFFECTEDROWS() . '|QUERYTIME:' . ($queryTime * 1000) . 'ms): ' . str_replace(array(chr(13), chr(10)), array('', ' '), $GLOBALS['last_sql']));
	} // END - if

	// Increment stats entry
	incrementStatsEntry('db_hits');

	// Return the result
	return $result;
}

// SQL num rows
function SQL_NUMROWS ($resource) {
	// Valid link resource?
	if (!SQL_IS_LINK_UP()) return FALSE;

	// Link is not up, no rows by default
	$lines = FALSE;

	// Is the result a valid resource?
	if (isset($GLOBALS['sql_numrows'][intval($resource)])) {
		// Use cache
		$lines = $GLOBALS['sql_numrows'][intval($resource)];
	} elseif (is_resource($resource)) {
		// Get the count of rows from database
		$lines = mysql_num_rows($resource);

		// Remember it in cache
		$GLOBALS['sql_numrows'][intval($resource)] = $lines;
	} else {
		// No resource given, please fix this
		reportBug(__FUNCTION__, __LINE__, 'No resource given! result[]=' . gettype($resource) . ',last_sql=' .  $GLOBALS['last_sql']);
	}

	// Return lines
	return $lines;
}

// SQL affected rows
function SQL_AFFECTEDROWS() {
	// Valid link resource?
	if (!SQL_IS_LINK_UP()) return FALSE;

	// Get affected rows
	$lines = mysql_affected_rows(SQL_GET_LINK());

	// Return it
	return $lines;
}

// SQL fetch row
function SQL_FETCHROW ($resource) {
	// Is a result resource set?
	if ((!is_resource($resource)) || (!SQL_IS_LINK_UP())) return FALSE;

	// Fetch the data and return it
	return mysql_fetch_row($resource);
}

// SQL fetch array
function SQL_FETCHARRAY ($resource) {
	// Is a result resource set?
	if ((!is_resource($resource)) || (!SQL_IS_LINK_UP())) return FALSE;

	// Load row from database
	$row = mysql_fetch_assoc($resource);

	// Return only arrays here
	if (is_array($row)) {
		// Return row
		return $row;
	} else {
		// Return a false, else some loops would go endless...
		return FALSE;
	}
}

// SQL result
function SQL_RESULT ($resource, $row, $field = '0') {
	// Is $resource valid?
	if ((!is_resource($resource)) || (!SQL_IS_LINK_UP())) return FALSE;

	// Run the result command
	$result = mysql_result($resource, $row, $field);

	// ... and return the result
	return $result;
}

// SQL connect
function SQL_CONNECT ($host, $login, $password, $F, $L) {
	// Try to connect
	$linkResource = mysql_connect($host, $login, $password) or SQL_ERROR($F, $L,  mysql_error());

	// Set the link resource
	if (is_resource($linkResource)) {
		/*
		 * A non-resource (boolean) may happen on installation phase which
		 * shall not be set here. Only valid link resources shall be set so
		 * SQL_IS_LINK_UP() will only return 'true' if there is really a
		 * working database link.
		 */
		SQL_SET_LINK($linkResource);
	} // END - if

	// Return the resource
	return $linkResource;
}

// SQL select database
function SQL_SELECT_DB ($dbName, $F, $L) {
	// Is there still a valid link? If not, skip it.
	if (!SQL_IS_LINK_UP()) return FALSE;

	// Return the result
	return mysql_select_db($dbName, SQL_GET_LINK()) or SQL_ERROR($F, $L,  mysql_error());
}

// SQL close link
function SQL_CLOSE ($F, $L) {
	if (!SQL_IS_LINK_UP()) {
		// Skip double close
		return FALSE;
	} // END - if

	// Close database link and forget the link
	$close = mysql_close(SQL_GET_LINK()) or SQL_ERROR($F, $L, mysql_error());

	// Close link
	SQL_SET_LINK(NULL);

	// Return the result
	return $close;
}

// SQL free result
function SQL_FREERESULT ($resource) {
	if ((!is_resource($resource)) || (!SQL_IS_LINK_UP())) {
		// Abort here
		return FALSE;
	} // END - if

	// Free result
	$res = mysql_free_result($resource);

	// And return that result of freeing it...
	return $res;
}

// SQL string escaping
function SQL_QUERY_ESC ($sqlString, $data, $F, $L, $run = TRUE, $strip = TRUE, $secure = TRUE) {
	// Link is there?
	if ((!SQL_IS_LINK_UP()) || (!is_array($data))) {
		// Link is down or data is not an array
		/* DEBUG: */ logDebugMessage($F, $L, 'SQL_IS_LINK_UP()=' . intval(SQL_IS_LINK_UP()) . ',data[]=' . gettype($data) . ',sqlString=' . $sqlString . ': ABORTING!');
		return FALSE;
	} // END - if

	// Init array for escape'd data with SQL string
	$dataSecured = array(
		'__sql_string' => $sqlString
	);

	// Escape all data
	foreach ($data as $key => $value) {
		$dataSecured[$key] = SQL_ESCAPE($value, $secure, $strip);
	} // END - foreach

	// Generate query
	$query = call_user_func_array('sprintf', $dataSecured);

	if ($run === TRUE) {
		// Run SQL query (default)
		return SQL_QUERY($query, $F, $L);
	} else {
		// Return secured string
		return $query;
	}
}

// Get id from last INSERT command
function SQL_INSERTID () {
	if (!SQL_IS_LINK_UP()) return FALSE;
	return mysql_insert_id();
}

// Escape a string for the database
function SQL_ESCAPE ($str, $secureString = TRUE, $strip = TRUE) {
	// Is there cache?
	if (!isset($GLOBALS['sql_escapes']['' . $str . ''])) {
		// Debug message
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'str=' . $str . ' - BEFORE!');

		// Prepare the string here
		$str = SQL_PREPARE_SQL_STRING($str);

		// Debug message
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'str=' . $str . ' - AFTER!');

		// Secure string first? (which is the default behaviour!)
		if ($secureString === TRUE) {
			// Debug message
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'str=' . $str . ',strip=' . intval($strip) . ' - BEFORE!');

			// Then do it here
			$str = secureString($str, $strip);

			// Debug message
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'str=' . $str . ',strip=' . intval($strip) . ' - AFTER!');
		} // END - if

		// Init (invalid) value
		$ret = '!INVALID!';

		if (!SQL_IS_LINK_UP()) {
			// Fall-back to escapeQuotes() when there is no link
			$ret = escapeQuotes($str);
		} elseif (function_exists('mysql_real_escape_string')) {
			// Debug message
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'str=' . $str);

			// The new and improved version
			$ret = mysql_real_escape_string($str, SQL_GET_LINK());

			// Debug message
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'str=' . $str . ',ret=' . $ret);
		} elseif (function_exists('mysql_escape_string')) {
			// The obsolete function
			$ret = mysql_escape_string($str, SQL_GET_LINK());
		} else {
			// If nothing else works, fall back to escapeQuotes() again
			$ret = escapeQuotes($str);
		}

		// Log message
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'str=' . $str . ',ret=' . $ret);

		// Cache result
		$GLOBALS['sql_escapes']['' . $str . ''] = $ret;
	} // END - if

	// Log message
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'str=' . $str . ',sql_escapes=' . $GLOBALS['sql_escapes']['' . $str . '']);

	// Return it
	return $GLOBALS['sql_escapes']['' . $str . ''];
}

// SELECT query string from table, columns and so on... ;-)
function SQL_RESULT_FROM_ARRAY ($table, $columns, $idRow, $id, $F, $L) {
	// Is columns an array?
	if (!is_array($columns)) {
		// No array
		reportBug(__FUNCTION__, __LINE__, sprintf("columns is not an array. %s != array, file=%s, line=%s",
			gettype($columns),
			basename($F),
			$L
		));

		// Abort here with 'false'
		return FALSE;
	} // END  - if

	// Is this is a simple array?
	if ((is_array($columns[0])) && (isset($columns[0]['column']))) {
		// Begin with SQL query
		$sql = 'SELECT ';

		// No, it comes from XML, so get it back from it
		$sql .= getSqlPartFromXmlArray($columns);

		// Finalize it
		$sql .= " FROM `{?_MYSQL_PREFIX?}_%s` WHERE `%s`='%s' LIMIT 1";
	} else {
		// Yes, prepare the SQL statement
		$sql = 'SELECT `' . implode('`, `', $columns) . "` FROM `{?_MYSQL_PREFIX?}_%s` WHERE `%s`='%s' LIMIT 1";
	}

	// Return the result
	return SQL_QUERY_ESC($sql,
		array(
			$table,
			$idRow,
			bigintval($id),
		), $F, $L
	);
}

// ALTER TABLE wrapper function
function SQL_ALTER_TABLE ($sql, $F, $L, $enableCodes = TRUE) {
	// Abort if link is down
	if (!SQL_IS_LINK_UP()) return FALSE;

	// This is the default result...
	$result = FALSE;

	// Determine index/fulltext/unique word
	$isAlterIndex = (
		(
			isInString('INDEX', $sql)
		) || (
			isInString('KEY', $sql)
		) || (
			isInString('FULLTEXT', $sql)
		) || (
			isInString('UNIQUE', $sql)
		)
	);

	// Extract table name
	$tableArray = explode(' ', $sql);
	$tableName = str_replace('`', '', $tableArray[2]);

	// Debug log
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'sql=' . $sql . ',tableName=' . $tableName . ',tableArray=<pre>' . print_r($tableArray, TRUE) . '</pre>,isAlterIndex=' . intval($isAlterIndex));

	// Shall we add/drop?
	if (((isInString('ADD', $sql)) || (isInString('DROP', $sql)) || (isInString('CHANGE', $sql))) && ($isAlterIndex === FALSE)) {
		// Try two columns, one should fix
		foreach (array(4,5) as $idx) {
			// If an entry is not set, abort here
			if (!isset($tableArray[$idx])) {
				// Debug log this
				logDebugMessage(__FUNCTION__, __LINE__, 'columnName=' . $columnName . ',idx=' . $idx . ',sql=' . $sql . ' is missing!');
				break;
			} // END - if

			// And column name as well
			$columnName = $tableArray[$idx];

			// Debug log
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'columnName=' . $columnName . ',idx=' . $idx . ',sql=' . $sql . ',hasZeroNums=' . intval(ifSqlColumnExists($tableName, $columnName)));

			// Is there no entry on ADD or an entry on DROP/CHANGE?
			if (((!ifSqlColumnExists($tableName, $columnName)) && (isInString('ADD', $sql))) || ((ifSqlColumnExists($tableName, $columnName)) && ((isInString('DROP', $sql)) || ((isInString('CHANGE', $sql)) && ($idx == 4) && ((!ifSqlColumnExists($tableName, $tableArray[5])) || ($columnName == $tableArray[5])))))) {
				// Do the query
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Executing: ' . $sql);
				$result = SQL_QUERY($sql, $F, $L, FALSE);

				// Skip further attempt(s)
				break;
			} elseif ((((ifSqlColumnExists($tableName, $columnName)) && (isInString('ADD', $sql))) || ((!ifSqlColumnExists($tableName, $columnName)) && ((isInString('DROP', $sql))) || (isInString('CHANGE', $sql)))) && ($columnName != 'KEY')) {
				// Abort here because it is alreay there
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Skipped: sql=' . $sql . ',columnName=' . $columnName . ',idx=' . $idx);
				break;
			} elseif ((!ifSqlColumnExists($tableName, $columnName)) && (isInString('DROP', $sql))) {
				// Abort here because we tried to drop a column which is not there (never created maybe)
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'No drop: ' . $sql);
				break;
			} elseif ($columnName != 'KEY') {
				// Something didn't fit, we better log it
				logDebugMessage(__FUNCTION__, __LINE__, 'Possible problem: ' . $sql . ',hasZeroNums=' . intval(ifSqlColumnExists($tableName, $columnName)) . '');
			}
		} // END - foreach
	} elseif ((getTableType() == 'InnoDB') && (isInString('FULLTEXT', $sql))) {
		// Skip this query silently because InnoDB does not understand fulltext indexes
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, sprintf("Skipped FULLTEXT: sql=%s,tableName=%s,hasZeroNums=%d,file=%s,line=%s", $sql, $tableName, intval((is_bool($result)) ? 0 : ifSqlColumnExists($columnName)), $F, $L));
	} elseif ($isAlterIndex === TRUE) {
		// And column name as well without backticks
		$keyName = str_replace('`', '', $tableArray[5]);
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'keyName=' . $keyName . ',tableArray=<pre>' . print_r($tableArray, TRUE) . '</pre>');

		// Is this "UNIQUE" or so? FULLTEXT has been handled the elseif() block above
		if (in_array(strtoupper($tableArray[4]), array('INDEX', 'UNIQUE', 'KEY', 'FULLTEXT'))) {
			// Init loop
			$begin = 1;
			$keyName = ',';
			while (isInString(',', $keyName)) {
				// Use last
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'keyName=' . $keyName . 'begin=' . $begin . ' - BEFORE');
				$keyName = str_replace('`', '', $tableArray[count($tableArray) - $begin]);
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'keyName=' . $keyName . 'begin=' . $begin . ' - BETWEEN');

				// Remove brackes
				$keyName = str_replace(array('(', ')'), array('', ''), $keyName);
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'keyName=' . $keyName . 'begin=' . $begin . ' - AFTER');

				// Continue
				$begin++;
			} // END while
		} // END - if

		// Shall we run it?
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'tableName=' . $tableName . ', tableArray[3]=' . $tableArray[3] . ',keyName=' . $keyName);
		if (($tableArray[3] == 'ADD') && (!isSqlTableIndexFound($tableName, $keyName))) {
			// Send it to the SQL_QUERY() function to add it
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'sql=' . $sql . ' - ADDING!');
			$result = SQL_QUERY($sql, $F, $L, $enableCodes);
		} elseif (($tableArray[3] == 'DROP') && (isSqlTableIndexFound($tableName, $keyName))) {
			// Send it to the SQL_QUERY() function to drop it
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'sql=' . $sql . ' - DROPPING!');
			$result = SQL_QUERY($sql, $F, $L, $enableCodes);
		} else {
			// Not executed
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Not executed: ' . $sql);
		}
	} else {
		// Other ALTER TABLE query
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, $sql);
		$result = SQL_QUERY($sql, $F, $L, $enableCodes);
	}

	// Return result
	return $result;
}

// Getter for SQL link
function SQL_GET_LINK () {
	// Init link
	$link = NULL;

	// Is it in the globals?
	if (isset($GLOBALS['sql_link'])) {
		// Then take it
		$link = $GLOBALS['sql_link'];
	} // END - if

	// Return it
	return $link;
}

// Setter for link
function SQL_SET_LINK ($link) {
	// Is this a resource or null?
	if ((ifFatalErrorsDetected()) && (isInstallationPhase())) {
		// This may happen in installation phase
		return;
	} elseif ((!is_resource($link)) && (!is_null($link))) {
		// This should never happen!
		reportBug(__FUNCTION__, __LINE__, sprintf("Type of link is not resource or null, type=%s", gettype($link)));
	} // END - if

	// Set it
	$GLOBALS['sql_link'] = $link;

	// Re-init cache
	$GLOBALS['is_sql_link_up'] = is_resource($link);
}

// Checks if the link is up
function SQL_IS_LINK_UP () {
	// Is there cached this?
	if (!isset($GLOBALS['is_sql_link_up'])) {
		// Determine it
		$GLOBALS['is_sql_link_up'] = is_resource(SQL_GET_LINK());
	} // END - if

	// Return the result
	return $GLOBALS['is_sql_link_up'];
}

// Wrapper function to make code more readable
function SQL_HASZERONUMS ($result) {
	// Just pass it through
	return (SQL_NUMROWS($result) === 0);
}

// Wrapper function to make code more readable
function SQL_HASZEROAFFECTED () {
	// Just pass it through
	return (SQL_AFFECTEDROWS() === 0);
}

// Private function to prepare the SQL query string
function SQL_PREPARE_SQL_STRING ($sqlString, $enableCodes = TRUE) {
	// Debug message
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'sqlString=' . $sqlString . ',enableCodes=' . intval($enableCodes) . ' - ENTERED!');

	// Is it already cached?
	if (!isset($GLOBALS['sql_strings']['' . $sqlString . ''])) {
		// Preserve escaping and compile URI codes+config+expression code
		$sqlString2 = FILTER_COMPILE_EXPRESSION_CODE(FILTER_COMPILE_CONFIG($sqlString));

		// Debug message
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'sqlString2=' . $sqlString2);

		// Do final compilation and revert {ESCAPE}
		$GLOBALS['sql_strings']['' . $sqlString . ''] = doFinalCompilation($sqlString2, FALSE, $enableCodes);
	} else {
		// Log message
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'sqlString=' . $sqlString . ' - CACHE!');
	}

	// Debug message
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'sqlString=' . $sqlString . ',enableCodes=' . intval($enableCodes) . ',sql_strings=' . $GLOBALS['sql_strings']['' . $sqlString . ''] . ' - EXIT!');

	// Return it
	return $GLOBALS['sql_strings']['' . $sqlString . ''];
}

// Creates a MySQL TIMESTAMP compatible string from given Uni* timestamp
function SQL_EPOCHE_TO_TIMESTAMP ($timestamp) {
	return generateDateTime($timestamp, 7);
}

// Log SQL errors to debug.log in installation phase or call reportBug()
function SQL_ERROR ($file, $line, $message) {
	// Remember plain error in last_sql_error
	$GLOBALS['last_sql_error'] = mysql_error();

	// Is there installation phase?
	if (isInstallationPhase()) {
		/*
		 * In installation phase, we don't want SQL errors abort e.g. connection
		 * tests, so just log it away.
		 */
		logDebugMessage($file, $line, $message);
	} else {
		// Regular mode, then call reportBug()
		reportBug($file, $line, $message);
	}
}

// Check if there is a SQL table created
function ifSqlTableExists ($tableName) {
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'tableName=' . $tableName . ' - ENTERED!');
	// Is there cache?
	if (!isset($GLOBALS[__FUNCTION__][$tableName])) {
		// Check if the table is there
		$result = SQL_QUERY_ESC("SHOW TABLES FROM `{?__DB_NAME?}` WHERE `Tables_in_{?__DB_NAME?}`='{?_MYSQL_PREFIX?}_%s'",
			array($tableName), __FILE__, __LINE__);

		// Is there an entry?
		$GLOBALS[__FUNCTION__][$tableName] = (SQL_NUMROWS($result) == 1);
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'tableName=' . $tableName . ',numRows=' . intval($GLOBALS[__FUNCTION__][$tableName]));
	} // END - if

	// Return cache
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'tableName=' . $tableName . ',result=' . intval($GLOBALS[__FUNCTION__][$tableName]) . ' - EXIT!');
	return $GLOBALS[__FUNCTION__][$tableName];
}

// Is a table column there?
function ifSqlColumnExists ($tableName, $columnName) {
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'tableName=' . $tableName . ',columnName=' . $columnName . ' - ENTERED!');
	// Is there cache?
	if (!isset($GLOBALS[__FUNCTION__][$tableName][$columnName])) {
		// And column name as well
		$columnName = str_replace('`', '', $columnName);

		// Get column information
		$result = SQL_QUERY_ESC("SHOW COLUMNS FROM `%s` LIKE '%s'",
			array(
				$tableName,
				$columnName
			), __FUNCTION__, __LINE__);

		// Determine it
		$GLOBALS[__FUNCTION__][$tableName][$columnName] = (!SQL_HASZERONUMS($result));
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'tableName=' . $tableName . ',columnName=' . $columnName . ',hasZeroNums=' . intval(SQL_HASZERONUMS($result)) . ',numRows=' . intval($GLOBALS[__FUNCTION__][$tableName][$columnName]));

		// Free result
		SQL_FREERESULT($result);
	} // END - if

	// Return cache
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'tableName=' . $tableName . ',columnName=' . $columnName . ',result=' . intval($GLOBALS[__FUNCTION__][$tableName][$columnName]) . ' - EXIT!');
	return $GLOBALS[__FUNCTION__][$tableName][$columnName];
}

// Checks depending on the mode if the index is there
function isSqlTableIndexFound ($tableName, $keyName) {
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'tableName=' . $tableName . ',columnName=' . $keyName . ' - ENTERED!');
	// Is there cache?
	if (!isset($GLOBALS[__FUNCTION__][$tableName][$keyName])) {
		// Show indexes
		$result = SQL_QUERY_ESC("SHOW INDEX FROM `%s`", array($tableName), __FUNCTION__, __LINE__);

		// The column is not found by default
		$GLOBALS[__FUNCTION__][$tableName][$keyName] = FALSE;

		// Walk through all
		while ($content = SQL_FETCHARRAY($result)) {
			// Add all entries for better caching behavior
			$GLOBALS[__FUNCTION__][$tableName][$content['Key_name']] = TRUE;
		} // END - while

		// Free result
		SQL_FREERESULT($result);
	} else {
		// Cache used
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'tableName=' . $tableName . ',columnName=' . $keyName . ',result=' . intval($GLOBALS[__FUNCTION__][$tableName][$keyName]) . ' - CACHE!');
	} // END - if

	// Return cache
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'tableName=' . $tableName . ',columnName=' . $keyName . ',result=' . intval($GLOBALS[__FUNCTION__][$tableName][$keyName]) . ' - EXIT!');
	return $GLOBALS[__FUNCTION__][$tableName][$keyName];
}

// Init database layer
function initDatabaseLayer () {
	// Set all required variables:
	$GLOBALS['last_sql_error'] = '';
}

// Get last SQL error
function getLastSqlError () {
	return $GLOBALS['last_sql_error'];
}

// Gets an array (or false if none is found) from all supported engines
function getArrayFromSupportedSqlEngines ($support = 'YES') {
	// Init array
	$engines = array();

	// This also worked, now we need to check if the selected database type is supported
	$result = SQL_QUERY('SHOW ENGINES', __FUNCTION__, __LINE__);

	// Are there entries? (Bad if not)
	if (SQL_NUMROWS($result) > 0) {
		// Load all and check for active entries
		while ($content = SQL_FETCHARRAY($result)) {
			// Is this supported?
			if (($support == 'ALL') || ($content['Support'] == $support)) {
				// Add it
				array_push($engines, $content);
			} elseif (isDebugModeEnabled()) {
				// Log it away in debug mode
				logDebugMessage(__FUNCTION__, __LINE__, 'Engine ' . $content['Engine'] . ' is not supported (' . $content['Supported'] . ' - ' . $support . ')');
			}
		} // END - if
	} else {
		// No engines! :(
		$engines = FALSE;
	}

	// Free result
	SQL_FREERESULT($result);

	// Return result
	return $engines;
}

// [EOF]
?>
